﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.DatastreamAnalysis {
  [StorableClass]
  [Item("Datastream", "Represents an item containing a data source and reading options for a certain data stream.")]
  [Creatable(CreatableAttribute.Categories.Problems)]
  public class Datastream : ParameterizedNamedItem, IDatastream {

    protected const string ProblemDataParameterName = "ProblemData";
    protected const string InitialSlidingWindowParameterName = "InitialSlidingWindowParameter";
    protected const string SlidingWindowSizeParameterName = "SlidingWindowSize";
    protected const string SlidingWindowStepWidthParameterName = "SlidingWindowStepWidth";
    protected const string SlidingWindowMovementIntervalParameterName = "SlidingWindowMovementInterval";

    #region parameter properites

    public IValueParameter<RegressionProblemData> ProblemDataParameter {
      get { return (IValueParameter<RegressionProblemData>) Parameters[ProblemDataParameterName]; }  
    }

    public IFixedValueParameter<IntRange> InitialSlidingWindowParameter {
      get { return (IFixedValueParameter<IntRange>) Parameters[InitialSlidingWindowParameterName]; }
    }

    public IFixedValueParameter<IntValue> SlidingWindowSizeParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[InitialSlidingWindowParameterName]; }
    }

    public IFixedValueParameter<IntValue> SlidingWindowStepWidthParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[SlidingWindowStepWidthParameterName]; }
    }

    public IFixedValueParameter<IntValue> SlidingWindowMovementIntervalParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[SlidingWindowMovementIntervalParameterName]; }
    }


    #endregion

    #region properties

    public RegressionProblemData ProblemData {
      get { return ProblemDataParameter.Value; }
      set {
        if(value == null) throw new ArgumentNullException("problemData", "The provided value for the problemData is null.");
        if (value == ProblemData) return;
        ProblemDataParameter.Value = value;
      }
    }

    public IntRange InitialSlidingWindow {
      get { return InitialSlidingWindowParameter.Value; }
    }

    public IntValue SlidingWindowSize {
      get { return SlidingWindowSizeParameter.Value; }
    }

    public IntValue SlidingWindowStepWidth {
      get { return SlidingWindowStepWidthParameter.Value; }
    }

    public IntValue SlidingWindowMovementInterval {
      get { return SlidingWindowMovementIntervalParameter.Value; }
    }

    #endregion

    #region internal properties

    public IntRange FitnessPartition { get; set; }

    public bool SlidingWindowMovementPossible {
      get {
        return ProblemData != null && ProblemData.Dataset != null &&
               FitnessPartition.End + SlidingWindowStepWidth.Value <= ProblemData.Dataset.Rows;
        // TODO: add external datastream, hub, etc to condition
      }
    }

    public bool SlidingWindowEvaluationPossible {
      get {
        return ProblemData != null && ProblemData.Dataset != null &&
               FitnessPartition.Size > 0;
      }
    }

    public bool MoveSlidingWindow() {
      if (ProblemData?.Dataset == null || ProblemData.Dataset.Rows == 0) return false;
      if (FitnessPartition.End + SlidingWindowStepWidth.Value > ProblemData.Dataset.Rows) return false;
      
      if (FitnessPartition.Size > SlidingWindowSize.Value) {
        FitnessPartition.Start += SlidingWindowStepWidth.Value;
      } else if (FitnessPartition.Size == SlidingWindowSize.Value) {
        FitnessPartition.Start += SlidingWindowStepWidth.Value;
        FitnessPartition.End += SlidingWindowStepWidth.Value;
      } else {
        FitnessPartition.End += SlidingWindowStepWidth.Value;
      }   
      return true;
    }

    #endregion

    #region constructors
    [StorableConstructor]
    protected Datastream(bool deserializing) : base(deserializing) { }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterParameterEvents();
    }

    protected Datastream(Datastream original, Cloner cloner) : base(original, cloner) {
      RegisterParameterEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Datastream(this, cloner);
    }

    public Datastream() {
      Parameters.Add(new ValueParameter<RegressionProblemData>(ProblemDataParameterName, "ProblemData for analysis with selected ensembles.", null));
      Parameters.Add(new FixedValueParameter<IntRange>(InitialSlidingWindowParameterName, "Initial sliding window boundaries", new IntRange(0, 1)));
      Parameters.Add(new FixedValueParameter<IntValue>(SlidingWindowSizeParameterName, "Sliding window size", new IntValue(1)));
      Parameters.Add(new FixedValueParameter<IntValue>(SlidingWindowStepWidthParameterName, "Sliding window step width", new IntValue(1)));
      Parameters.Add(new FixedValueParameter<IntValue>(SlidingWindowMovementIntervalParameterName, "Sliding window movement interval (milliseconds)", new IntValue(1000)));
      InitializeState();
    }
    #endregion


    public void InitializeState() {
      if (ProblemData == null || ProblemData.Dataset == null || ProblemData.Dataset.Rows == 0) {
        FitnessPartition = new IntRange(0, 0);
      } else {
        FitnessPartition = (IntRange)InitialSlidingWindow.Clone();
        if (FitnessPartition.End > ProblemData.Dataset.Rows) {
          FitnessPartition.End = ProblemData.Dataset.Rows;
        }
      }
    }

    public event EventHandler ProblemDataChanged;


    private void RegisterParameterEvents() {
      ProblemDataParameter.ValueChanged += new EventHandler(ProblemData_ValueChanged);
    }

    private void ProblemData_ValueChanged(object sender, EventArgs e) {
      if (e == null) return;

      DatastreamAnalysisUtil.RaiseEvent(this, ProblemDataChanged);
    }

    

  }
}
